"""This module contains the `SRIMDB` class."""

# pylint: disable=too-many-lines
# pylint: disable=protected-access
import os
import platform
import sqlite3
import subprocess
import threading
import time
import traceback
import warnings
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import Callable, Optional, Union

import numpy as np
from mpi4py import MPI

from irradiapy import dtypes, materials
from irradiapy.srimpy.ofiles.backscat import Backscat
from irradiapy.srimpy.ofiles.collision import Collision
from irradiapy.srimpy.ofiles.e2recoil import E2Recoil
from irradiapy.srimpy.ofiles.ioniz import Ioniz
from irradiapy.srimpy.ofiles.lateral import Lateral
from irradiapy.srimpy.ofiles.novac import Novac
from irradiapy.srimpy.ofiles.phonon import Phonon
from irradiapy.srimpy.ofiles.range import Range
from irradiapy.srimpy.ofiles.range3d import Range3D
from irradiapy.srimpy.ofiles.sputter import Sputter
from irradiapy.srimpy.ofiles.subcollision import Subcollision
from irradiapy.srimpy.ofiles.transmit import Transmit
from irradiapy.srimpy.ofiles.trimdat import Trimdat
from irradiapy.srimpy.ofiles.vacancy import Vacancy
from irradiapy.srimpy.target.element import Element
from irradiapy.srimpy.target.layer import Layer
from irradiapy.srimpy.target.target import Target

platform = platform.system()
if platform == "Windows":
    import pygetwindow
    import pywinauto
else:
    if MPI.COMM_WORLD.Get_rank() == 0:
        warnings.warn(
            ("SRIM subpackage only works for Windows. " f"'{platform}' not supported.")
        )


@dataclass(kw_only=True)
class SRIMDB(sqlite3.Connection):
    """Base class for running SRIM calculations and storing the output data in a SQLite database.

    Attributes
    ----------
    path_db : Path
        Output database path.
    target : Target, optional
        SRIM target.
    calculation : str, optional
        SRIM calculation.
    check_interval : float
        Interval to check for SRIM window/popups.
    srim_path : Path
        Where all SRIM output files are.
        If given, it will automatically add all those files into the database.
    con : sqlite3.Connection
        Database connection.
    backscat : Backscat
        Class storing `BACKSCAT.txt` data.
    e2recoil : E2Recoil
        Class storing `E2RECOIL.txt` data.
    ioniz : Ioniz
        Class storing `IONIZ.txt` data.
    lateral : Lateral
        Class storing `LATERAL.txt` data.
    phonon : Phonon
        Class storing `PHONON.txt` data.
    range : Range
        Class storing `RANGE.txt` data.
    range3d : Range3D
        Class storing `RANGE_3D.txt` data.
    sputter : Sputter
        Class storing `SPUTTER.txt` data.
    transmit : Transmit
        Class storing `TRANSMIT.txt` data.
    trimdat : Trimdat
        Class storing `TRIM.DAT` data.
    vacancy : Vacancy
        Class storing `VACANCY.txt` data.
    """

    path_db: Path
    calculation: Optional[str] = None
    target: Optional[Target] = None
    check_interval: float = 0.2
    seed: int = 0
    reminders: int = 0
    plot_type: int = 5
    xmin: float = 0.0
    xmax: float = 0.0
    do_ranges: int = 1
    do_backscatt: int = 1
    do_transmit: int = 1
    do_sputtered: int = 1
    do_collisions: int = 1
    exyz: float = 0.0
    bragg: int = 1
    autosave: int = 0

    def __post_init__(self) -> None:
        """Initializes the `SRIMDB` object.

        Parameters
        ----------
        path_db : Path
            Output database path.
        target : Target, optional
            SRIM target. Do not provide this argument for read only. By default None.
        calculation : Calculation, optional
            SRIM calculation. Do not provide this argument for read only. By default None.
        check_interval : float
            Interval to check for SRIM window/popups.
        """
        super().__init__(self.path_db)
        self.backscat = Backscat(self)
        self.e2recoil = E2Recoil(self)
        self.ioniz = Ioniz(self)
        self.lateral = Lateral(self)
        self.phonon = Phonon(self)
        self.range = Range(self)
        self.range3d = Range3D(self)
        self.sputter = Sputter(self)
        self.transmit = Transmit(self)
        self.trimdat = Trimdat(self)
        self.vacancy = Vacancy(self)
        self.collision = Collision(self)
        self.subcollision = Subcollision(self)

        if self.calculation not in ["quick", "full", "mono", None]:
            raise ValueError("Invalid calculation mode.")

        self.nions = self.get_nions()
        if self.target and self.calculation and not self.table_exists("calculation"):
            self.save_target_calculation()
        elif (
            not self.target
            and not self.calculation
            and self.table_exists("calculation")
        ):
            self.target = self.load_target_calculation()
        elif (self.target and not self.calculation) or (
            not self.target and self.calculation
        ):
            raise ValueError(
                "Both `target` and `calculation` must be provided or None."
            )

        if self.calculation == "quick":
            self._filter_subcollisions_logic = self.__filter_subcollisions_logic_qc
        else:
            self.novac = Novac(self)
            self._filter_subcollisions_logic = self.__filter_subcollisions_logic_fc

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        exc_traceback: Optional[TracebackType] = None,
    ) -> bool:
        """Exit the runtime context related to this object."""
        self.close()
        return False

    def save_target_calculation(self) -> None:
        """Saves the target and calculation parameters into the database."""
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE IF NOT EXISTS layers("
                "layer_numb INTEGER, width REAL,"
                "phase INTEGER, density REAL, bragg INTEGER)"
            )
        )
        cur.execute(
            (
                "CREATE TABLE IF NOT EXISTS elements("
                "layer_numb INTEGER, stoich REAL, symbol TEXT,"
                "atomic_number INTEGER, atomic_mass REAL,"
                "e_d REAL, e_l REAL, e_s REAL)"
            )
        )
        for i, layer in enumerate(self.target.layers):
            cur.execute(
                (
                    "INSERT INTO layers"
                    "(layer_numb, width, phase, density, bragg)"
                    "VALUES(?, ?, ?, ?, ?)"
                ),
                [i, layer.width, layer.phase, layer.density, layer.bragg],
            )
            for j, element in enumerate(layer.elements):
                cur.execute(
                    (
                        "INSERT INTO elements"
                        "(layer_numb, stoich, symbol, atomic_number, atomic_mass, e_d, e_l, e_s)"
                        "VALUES(?, ?, ?, ?, ?, ?, ?, ?)"
                    ),
                    [
                        i,
                        layer.stoichs[j],
                        element.symbol,
                        element.atomic_number,
                        element.atomic_mass,
                        element.e_d,
                        element.e_l,
                        element.e_s,
                    ],
                )
        cur.execute(
            (
                "CREATE TABLE IF NOT EXISTS calculation("
                "mode INTEGER, seed INTEGER, reminders INTEGER, plot_type INTEGER,"
                "xmin REAL, xmax REAL, ranges INTEGER, backscatt INTEGER,"
                "transmit INTEGER, sputtered INTEGER, collisions INTEGER, exyz REAL,"
                "bragg INTEGER, autosave INTEGER)"
            )
        )
        if self.calculation == "quick":
            calculation = 4
        elif self.calculation == "full":
            calculation = 5
        else:
            calculation = 7
        cur.execute(
            (
                "INSERT INTO calculation"
                "(mode, seed, reminders, plot_type, xmin, xmax, ranges, backscatt,"
                "transmit, sputtered, collisions, exyz, bragg, autosave)"
                "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
            ),
            [
                calculation,
                self.seed,
                self.reminders,
                self.plot_type,
                self.xmin,
                self.xmax,
                self.do_ranges,
                self.do_backscatt,
                self.do_transmit,
                self.do_sputtered,
                self.do_collisions,
                self.exyz,
                self.bragg,
                self.autosave,
            ],
        )
        cur.close()
        self.commit()

    def load_target_calculation(self) -> Target:
        """Loads the target and calculation parameters from the database."""
        cur = self.cursor()
        cur.execute("SELECT * FROM layers")
        db_layers = list(cur.fetchall())
        layers = []
        for db_layer in db_layers:
            cur.execute(("SELECT * FROM elements " f"WHERE layer_numb = {db_layer[0]}"))
            db_elements = list(cur.fetchall())
            elements = []
            for db_element in db_elements:
                print(db_element)
                elements.append(
                    Element(
                        db_element[2],
                        db_element[3],
                        db_element[4],
                        db_element[5],
                        db_element[6],
                        db_element[7],
                    )
                )
            stoichs = [db_element[1] for db_element in db_elements]
            layers.append(
                Layer(
                    db_layer[1],
                    db_layer[2],
                    db_layer[3],
                    elements,
                    stoichs,
                    db_layer[4],
                )
            )
        cur.execute("SELECT * FROM calculation")
        db_calculation = cur.fetchone()
        cur.close()
        self.seed = db_calculation[1]
        self.reminders = db_calculation[2]
        self.plot_type = db_calculation[3]
        self.xmin = db_calculation[4]
        self.xmax = db_calculation[5]
        self.do_ranges = db_calculation[6]
        self.do_backscatt = db_calculation[7]
        self.do_transmit = db_calculation[8]
        self.do_sputtered = db_calculation[9]
        self.do_collisions = db_calculation[10]
        self.exyz = db_calculation[11]
        self.bragg = db_calculation[12]
        self.autosave = db_calculation[13]
        mode = db_calculation[0]
        if mode == 4:
            self.calculation = "quick"
        elif mode == 5:
            self.calculation = "full"
        else:
            self.calculation = "mono"
        cur.close()
        return Target(layers)

    def optimize(self) -> None:
        """Optimize the SQLite database.

        This method performs two operations to optimize the database:
        1. Executes the "PRAGMA optimize" command to analyze and optimize the database.
        2. Executes the "VACUUM" command to rebuild the database file,
        repacking it into a minimal amount of disk space.
        """
        cur = self.cursor()
        cur.execute("PRAGMA optimize")
        cur.execute("VACUUM")
        cur.close()

    def table_exists(self, table_name: str) -> bool:
        """Checks if the given table already exists in the database.

        Parameters
        ----------
        table_name : str
            Table's name to check.

        Returns
        -------
        bool
            Whether the table already exists or not.
        """
        cur = self.cursor()
        cur.execute(
            (
                "SELECT COUNT(*) FROM sqlite_master WHERE type='table'"
                f"AND name='{table_name}'"
            )
        )
        result = cur.fetchone()[0]
        cur.close()
        return bool(result)

    def get_nions(self) -> int:
        """Gets the number of ions in the simulation."""
        if self.table_exists("trimdat"):
            cur = self.cursor()
            cur.execute("SELECT COUNT(1) FROM trimdat")
            nions = cur.fetchone()[0]
            cur.close()
            return nions
        return 0

    def append_backscat(self, backscat_path: Path) -> None:
        """Appends `BACKSCAT.txt` into the database.

        Parameters
        ----------
        backscat_path : Path
            `BACKSCAT.txt` path.
        """
        self.backscat.process_file(backscat_path)

    def append_e2recoil(self, e2recoil_path: Path) -> None:
        """Appends `E2RECOIL.txt` into the database.

        Parameters
        ----------
        e2recoil_path : Path
            `E2RECOIL.txt` path.
        """
        self.e2recoil.process_file(e2recoil_path)

    def append_ioniz(self, ioniz_path: Path) -> None:
        """Appends `IONIZ.txt` into the database.

        Parameters
        ----------
        ioniz_path : Path
            `IONIZ.txt` path.
        """
        self.ioniz.process_file(ioniz_path)

    def append_lateral(self, lateral_path: Path) -> None:
        """Appends `LATERAL.txt` into the database.

        Parameters
        ----------
        lateral_path : Path
            `LATERAL.txt` path.
        """
        self.lateral.process_file(lateral_path)

    def append_phonon(self, phonon_path: Path) -> None:
        """Appends `PHONON.txt` into the database.

        Parameters
        ----------
        phonon_path : Path
            `PHONON.txt` path.
        """
        self.phonon.process_file(phonon_path)

    def append_range(self, range_path: Path) -> None:
        """Appends `RANGE.txt` into the database.

        Parameters
        ----------
        range_path : Path
            `RANGE.txt` path.
        """
        self.range.process_file(range_path)

    def append_range3d(self, range3d_path: Path) -> None:
        """Appends `RANGE_3D.txt` into the database.

        Parameters
        ----------
        range3d_path : Path
            `RANGE_3D.txt` path.
        """
        self.range3d.process_file(range3d_path)

    def append_sputter(self, sputter_path: Path) -> None:
        """Appends `SPUTTER.txt` into the database.

        Parameters
        ----------
        sputter_path : Path
            `SPUTTER.txt` path.
        """
        self.sputter.process_file(sputter_path)

    def append_transmit(self, transmit_path: Path) -> None:
        """Appends `TRANSMIT.txt` into the database.

        Parameters
        ----------
        transmit_path : Path
            `TRANSMIT.txt` path.
        """
        self.transmit.process_file(transmit_path)

    def append_trimdat(self, trimdat_path: Path) -> None:
        """Appends `TRIM.DAT` into the database.

        Parameters
        ----------
        trimdat_path : Path
            `TRIM.DAT` path.
        """
        self.trimdat.process_file(trimdat_path)
        self.nions = self.get_nions()

    def append_vacancy(self, vacancy_path: Path) -> None:
        """Appends `VACANCY.txt` into the database.

        Parameters
        ----------
        vacancy_path : Path
            `VACANCY.txt` path.
        """
        self.vacancy.process_file(vacancy_path)

    def append_subcollision(self, collision_path: Path) -> None:
        """Appends currect iteration `COLLISON.txt` into the database.

        Parameters
        ----------
        collision_path : Path
            `COLLISON.txt` path.
        """
        self.subcollision.process_file(collision_path)

    def append_novac(self, novac_path: Path) -> None:
        """Appends `NOVAC.txt` into the database.

        Parameters
        ----------
        novac_path : Path
            `NOVAC.txt` path.
        """
        self.novac.process_file(novac_path)

    def merge(
        self,
        srimdb2: "SRIMDB",
        backscat: bool = True,
        e2recoil: bool = True,
        ioniz: bool = True,
        lateral: bool = True,
        phonon: bool = True,
        range3d: bool = True,
        range_: bool = True,
        sputter: bool = True,
        transmit: bool = True,
        vacancy: bool = True,
        collision: bool = True,
        trimdat: bool = True,
        novac: bool = True,
    ) -> None:
        """Merges two databases.

        Parameters
        ----------
        srimdb2 : SRIMDBIter
            SRIM database to merge.
        backscat : bool, optional
            Merge backscattering data.
        e2recoil : bool, optional
            Merge energy to recoil data.
        ioniz : bool, optional
            Merge ionization data.
        lateral : bool, optional
            Merge lateral data.
        phonon : bool, optional
            Merge phonon data.
        range3d : bool, optional
            Merge 3D range data.
        range_ : bool, optional
            Merge range data.
        sputter : bool, optional
            Merge sputtering data.
        transmit : bool, optional
            Merge transmission data.
        vacancy : bool, optional
            Merge vacancy data.
        collision : bool, optional
            Merge collision data.
        trimdat : bool, optional
            Merge TRIMDAT data.
        novac : bool, optional
            Merge NOVAC data.
        """
        if backscat:
            self.backscat.merge(srimdb2)
        if e2recoil:
            self.e2recoil.merge(srimdb2)
        if ioniz:
            self.ioniz.merge(srimdb2)
        if lateral:
            self.lateral.merge(srimdb2)
        if phonon:
            self.phonon.merge(srimdb2)
        if range3d:
            self.range3d.merge(srimdb2)
        if range_:
            self.range.merge(srimdb2)
        if sputter:
            self.sputter.merge(srimdb2)
        if transmit:
            self.transmit.merge(srimdb2)
        if vacancy:
            self.vacancy.merge(srimdb2)
        if collision:
            self.collision.merge(srimdb2)
        if self.calculation in ["full", "mono"] and novac:
            self.novac.merge(srimdb2)
        if trimdat:
            self.trimdat.merge(srimdb2)
        self.optimize()

    def generate_trimin(
        self,
        srim_dir: Path,
        atomic_numbers: np.ndarray,
        energies: np.ndarray,
        target: Target,
    ) -> None:
        """Generates `TRIM.IN` file."""
        nions = len(atomic_numbers)
        atomic_mass = materials.get_mass_number_by_atomic_number(atomic_numbers[0])
        energy = np.ceil(energies.max()) / 1e3
        if self.calculation == "quick":
            calculation = 4
        elif self.calculation == "full":
            calculation = 5
        else:
            calculation = 7
        trimin_path = srim_dir / "TRIM.IN"
        with open(trimin_path, "w", encoding="utf-8") as file:
            file.write("TRIM.IN file generated by irradiapy.\n")
            file.write(
                (
                    "ion Z, A, energy, angle, number of ions, Bragg correction, autosave\n"
                    f"{atomic_numbers[0]} {atomic_mass} {energy} 0 "
                    f"{nions} {self.bragg} {self.autosave}\n"
                )
            )
            file.write(
                (
                    "calculation, seed, reminders\n"
                    f"{calculation} {self.seed} {self.reminders}\n"
                )
            )
            file.write(
                (
                    "Diskfiles (0=no,1=yes): Ranges, Backscatt, Transmit, Sputtered, "
                    "Collisions(1=Ion;2=Ion+Recoils), Special EXYZ.txt file\n"
                    f"{self.do_ranges} {self.do_backscatt} {self.do_transmit} {self.do_sputtered} "
                    f"{self.do_collisions} {self.exyz}\n"
                )
            )
            file.write(target.trimin_description() + "\n")
            file.write(
                (
                    "PlotType (0-5); Plot Depths: Xmin, Xmax(Ang.) [=0 0 for Viewing Full Target]\n"
                    f"{self.plot_type} {self.xmin} {self.xmax}\n"
                )
            )
            file.write(target.trimin_target_elements())
            file.write(target.trimin_target_layers() + "\n")
            file.write(target.trimin_phases() + "\n")
            file.write(target.trimin_bragg() + "\n")
            file.write(target.trimin_displacement() + "\n")
            file.write(target.trimin_lattice() + "\n")
            file.write(target.trimin_surface() + "\n")
            file.write("Stopping Power Version\n0\n")

    @staticmethod
    def generate_trimdat(
        ofile_dir: Path,
        atomic_numbers: np.ndarray,
        energies: np.ndarray,
        depths: Optional[np.ndarray] = None,
        ys: Optional[np.ndarray] = None,
        zs: Optional[np.ndarray] = None,
        cosxs: Optional[np.ndarray] = None,
        cosys: Optional[np.ndarray] = None,
        coszs: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Generates `TRIM.DAT` file.

        Parameters
        ----------
        ofile_dir : Path
            Output file directory.
        atomic_numbers : dtypes.INT_CHECK
            Atomic numbers.
        energies : dtypes.FLOAT_CHECK
            Energies.
        depths : dtypes.FLOAT_CHECK, optional
            Depths.
        ys : dtypes.FLOAT_CHECK, optional
            Y positions.
        zs : dtypes.FLOAT_CHECK, optional
            Z positions.
        cosxs : dtypes.FLOAT_CHECK, optional
            X directions.
        cosys : dtypes.FLOAT_CHECK, optional
            Y directions.
        coszs : dtypes.FLOAT_CHECK, optional
            Z directions.

        Returns
        -------
        np.ndarray
            `TRIM.DAT` data.
        """
        ofile_path = ofile_dir / "TRIM.DAT"
        nions = atomic_numbers.size
        if depths is None:
            depths = np.zeros(nions)
        if ys is None:
            ys = np.zeros(nions)
        if zs is None:
            zs = np.zeros(nions)
        if cosxs is None:
            cosxs = np.ones(nions)
        if cosys is None:
            cosys = np.zeros(nions)
        if coszs is None:
            coszs = np.zeros(nions)
        names = np.array([f"{i:06d}" for i in range(1, nions + 1)], dtype=str)
        with open(ofile_path, "w", encoding="utf-8") as file:
            file.write("\n" * 9)
            file.write("Name atomic_number E x y z cosx cosy cosz\n")
            for i in range(nions):
                file.write(
                    (
                        f"{names[i]} {atomic_numbers[i]} {energies[i]} {depths[i]} {ys[i]} "
                        f"{zs[i]} {cosxs[i]} {cosys[i]} {coszs[i]}\n"
                    )
                )
        data = np.empty(nions, dtype=dtypes.trimdat)
        for i in range(nions):
            data[i]["name"] = names[i]
            data[i]["atomic_number"] = atomic_numbers[i]
            data[i]["energy"] = energies[i]
            data[i]["pos"] = np.array([depths[i], ys[i], zs[i]])
            data[i]["dir"] = np.array([cosxs[i], cosys[i], coszs[i]])
        return data

    @staticmethod
    def __generate_trimauto(srim_dir) -> None:
        """Generates `TRIMAUTO` file."""
        with open(srim_dir / "TRIMAUTO", "w", encoding="utf-8") as file:
            file.write("1\n\nCheck TRIMAUTO.txt for details.\n")

    def minimize_and_handle_popup(self):
        """Minimizes the SRIM window and handles the end of calculation popup."""
        window_title = "SRIM-2013.00"
        popup_title = "End of TRIM.DAT calculation"
        if platform == "Windows":
            # Minimize window
            while True:
                windows = pygetwindow.getWindowsWithTitle(window_title)
                if windows:
                    window = windows[0]
                    app = pywinauto.Application().connect(handle=window._hWnd)
                    app.window(handle=window._hWnd).minimize()
                    break
                time.sleep(self.check_interval)
            # Dismiss popup (quick-calculation does not have this popup)
            if self.calculation != "quick":
                while True:
                    popups = pygetwindow.getWindowsWithTitle(popup_title)
                    if popups:
                        popup = popups[0]
                        app = pywinauto.Application().connect(handle=popup._hWnd)
                        app.window(handle=popup._hWnd).send_keystrokes("{ENTER}")
                        break
                    time.sleep(self.check_interval)
        elif platform == "Linux":
            # Minimize window
            while True:
                windows = pygetwindow.getWindowsWithTitle(window_title)
                if windows:
                    window = windows[0]
                    window_id = window._hWnd  # pylint: disable=protected-access
                    subprocess.run(
                        ["xdotool", "windowminimize", str(window_id)], check=True
                    )
                    break
                time.sleep(self.check_interval)
            # Dismiss popup (quick-calculation does not have this popup)
            if self.calculation != "quick":
                while True:
                    popups = pygetwindow.getWindowsWithTitle(popup_title)
                    if popups:
                        popup = popups[0]
                        popup_id = popup._hWnd  # pylint: disable=protected-access
                        subprocess.run(
                            ["xdotool", "windowactivate", str(popup_id)], check=True
                        )
                        subprocess.run(["xdotool", "key", "Return"], check=True)
                        break
                    time.sleep(self.check_interval)

    def run(
        self,
        srim_dir: Path,
        criterion: Callable,
        atomic_numbers: np.ndarray,
        energies: np.ndarray,
        remove_offsets: bool,
        depths: Optional[np.ndarray] = None,
        ys: Optional[np.ndarray] = None,
        zs: Optional[np.ndarray] = None,
        cosxs: Optional[np.ndarray] = None,
        cosys: Optional[np.ndarray] = None,
        coszs: Optional[np.ndarray] = None,
        iter_max: Optional[int] = None,
        ignore_32bit_warning: bool = True,
    ) -> None:
        """Runs the SRIM simulation.

        Parameters
        ----------
        criterion : Callable
            Criterion to repeat calculation, must return False to repeat calculation.
            Its signature is:
            `criterion(nion, energy, depth, y, z, se, atom_hit,
            pka_e, target_disp)`.
            Recommended to be defined as `def criterion(**kwargs: dict) -> bool:`.
        atomic_numbers : dtypes.INT_CHECK
            Ion atomic numbers.
        energies : dtypes.FLOAT_CHECK
            Ion energies.
        remove_offsets : bool
            Whether to remove initial depth offsets or not.
        depths : dtypes.FLOAT_CHECK, optional
            Ion initial depths.
        ys : dtypes.FLOAT_CHECK, optional
            Ion initial y positions.
        zs : dtypes.FLOAT_CHECK, optional
            Ion initial z positions.
        cosxs : dtypes.FLOAT_CHECK, optional
            Ion initial x directions.
        cosys : dtypes.FLOAT_CHECK, optional
            Ion initial y directions.
        coszs : dtypes.FLOAT_CHECK, optional
            Ion initial z directions.
        iter_max : int, optional
            Maximum number of iterations, by default None.
        """
        if ignore_32bit_warning:
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="32-bit application should be automated using 32-bit Python",
            )

        if self.table_exists("collision"):
            raise RuntimeError(
                (
                    f"The database {self.path_db} is already populated "
                    "with data from another simulation, use another one."
                )
            )
        self.collision.create_table()
        cur = self.cursor()

        # First iteration
        nions = len(atomic_numbers)
        nsubcollisions = np.ones(nions, dtype=int)
        nsubcollisions0 = nsubcollisions.copy()
        trimdat = self._run_iter(
            srim_dir,
            self.target,
            atomic_numbers,
            energies,
            depths,
            ys,
            zs,
            cosxs,
            cosys,
            coszs,
        )
        self.__append_output(srim_dir)
        (
            nsubcollisions,
            atomic_numbers,
            energies,
            depths,
            ys,
            zs,
            cosxs,
            cosys,
            coszs,
        ) = self._filter_subcollisions(
            srim_dir, cur, nions, trimdat, nsubcollisions0, criterion
        )
        nsubcollisions_total = nsubcollisions.sum()

        niter = 1
        while nsubcollisions_total and niter != iter_max:
            niter += 1
            nsubcollisions0 = nsubcollisions.copy()
            trimdat = self._run_iter(
                srim_dir,
                self.target,
                atomic_numbers,
                energies,
                depths,
                ys,
                zs,
                cosxs,
                cosys,
                coszs,
            )
            (
                nsubcollisions,
                atomic_numbers,
                energies,
                depths,
                ys,
                zs,
                cosxs,
                cosys,
                coszs,
            ) = self._filter_subcollisions(
                srim_dir, cur, nions, trimdat, nsubcollisions0, criterion
            )
            nsubcollisions_total = nsubcollisions.sum()

        # Reordering database might reduce I/O operations filtered by ion number
        # PKAs are not sequentally ordered
        cur.execute(
            "CREATE TABLE collision0 AS SELECT * FROM collision ORDER BY ion_numb"
        )
        cur.execute("DROP TABLE collision")
        cur.execute("ALTER TABLE collision0 RENAME TO collision")
        cur.execute(
            "CREATE INDEX ionRecoilEenrgyIdx ON collision(ion_numb, recoil_energy)"
        )
        self.commit()
        cur.close()
        if remove_offsets:
            self.__remove_offsets()
        self.optimize()

        if ignore_32bit_warning:
            warnings.filterwarnings(
                "default",
                category=UserWarning,
                message="32-bit application should be automated using 32-bit Python",
            )

    def _run_iter(
        self,
        srim_dir: Path,
        target: Target,
        atomic_numbers: np.ndarray,
        energies: np.ndarray,
        depths: Optional[np.ndarray] = None,
        ys: Optional[np.ndarray] = None,
        zs: Optional[np.ndarray] = None,
        cosxs: Optional[np.ndarray] = None,
        cosys: Optional[np.ndarray] = None,
        coszs: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Runs SRIM for a single iteration.

        Parameters
        ----------
        target : Target
            Target material.
        calculation : Calculation
            Calculation parameters.
        atomic_numbers : dtypes.INT_CHECK
            Ion atomic numbers.
        energies : dtypes.FLOAT_CHECK
            Ion energies.
        depths : dtypes.FLOAT_CHECK, optional
            Ion initial depths.
        ys : dtypes.FLOAT_CHECK, optional
            Ion initial y positions.
        zs : dtypes.FLOAT_CHECK, optional
            Ion initial z positions.
        cosxs : dtypes.FLOAT_CHECK, optional
            Ion initial x directions.
        cosys : dtypes.FLOAT_CHECK, optional
            Ion initial y directions.
        coszs : dtypes.FLOAT_CHECK, optional
            Ion initial z directions.

        Returns
        -------
        dtypes.trimdat
            `TRIM.DAT` data.
        """
        self.__generate_trimauto(srim_dir)
        self.generate_trimin(srim_dir, atomic_numbers, energies, target)
        trimdat = self.generate_trimdat(
            srim_dir,
            atomic_numbers,
            energies,
            depths=depths,
            ys=ys,
            zs=zs,
            cosxs=cosxs,
            cosys=cosys,
            coszs=coszs,
        )
        # Run SRIM
        print(f"Running {len(atomic_numbers)} SRIM ions")
        try:
            window_thread = threading.Thread(target=self.minimize_and_handle_popup)
            window_thread.start()
            curr_dir = os.getcwd()
            os.chdir(srim_dir)
            subprocess.check_call([Path("./TRIM.exe")])
            os.chdir(curr_dir)
            window_thread.join()
        except subprocess.CalledProcessError as e:
            print(traceback.format_exc())
            print(f"An error occurred while running the subprocess: {e}")
        return trimdat

    def __append_output(self, srim_dir: Path) -> None:
        """Appends SRIM output files into the database.

        Note
        ----
        Although reading and saving information from the TRIM.DAT and TRIM.IN files
        instead of directly using the Python objects provided as input to SRIMDB.run
        may seem redundant, this approach is maintained to preserve the ability to read
        SRIM configuration files. The resulting performance loss is relatively minor.
        """
        self.append_backscat(srim_dir / "SRIM Outputs/BACKSCAT.txt")
        self.append_e2recoil(srim_dir / "E2RECOIL.txt")
        self.append_ioniz(srim_dir / "IONIZ.txt")
        self.append_lateral(srim_dir / "LATERAL.txt")
        self.append_phonon(srim_dir / "PHONON.txt")
        self.append_range3d(srim_dir / "SRIM Outputs/RANGE_3D.txt")
        self.append_range(srim_dir / "RANGE.txt")
        self.append_sputter(srim_dir / "SRIM Outputs/SPUTTER.txt")
        self.append_transmit(srim_dir / "SRIM Outputs/TRANSMIT.txt")
        self.append_trimdat(srim_dir / "TRIM.DAT")
        self.append_vacancy(srim_dir / "VACANCY.txt")
        if self.calculation in ["full", "mono"]:
            self.append_novac(srim_dir / "NOVAC.txt")

    def _get_dir(self, pos0: np.ndarray, pos: np.ndarray) -> np.ndarray:
        """Gets the direction from two positions.

        Parameters
        pos0 : np.ndarray
            Initial position.
        pos : np.ndarray
            Final position.
        """
        diff = pos - pos0
        direction = diff / np.linalg.norm(diff)
        return direction

    def __filter_subcollisions_logic_qc(self, subcollision_data: tuple) -> dict:
        """Convert subcollision data into dictionary for
        later handling for Quick-Calculation mode.

        Parameters
        ----------
        subcollision_data : tuple
            Subcollision data.
        """
        _, energy, depth, y, z, se, atom_hit, recoil_energy, target_disp = (
            subcollision_data
        )
        return {
            "energy": energy,
            "depth": depth,
            "y": y,
            "z": z,
            "se": se,
            "atom_hit": atom_hit,
            "recoil_energy": recoil_energy,
            "target_disp": target_disp,
        }

    def __filter_subcollisions_logic_fc(self, subcollision_data: tuple) -> dict:
        """Convert subcollision data into dictionary for
        later handling for Full-Calculation mode.

        Parameters
        ----------
        subcollision_data : tuple
            Subcollision data.
        """
        (
            _,
            energy,
            depth,
            y,
            z,
            se,
            atom_hit,
            recoil_energy,
            target_disp,
            target_vac,
            target_replac,
            target_inter,
        ) = subcollision_data
        return {
            "energy": energy,
            "depth": depth,
            "y": y,
            "z": z,
            "se": se,
            "atom_hit": atom_hit,
            "recoil_energy": recoil_energy,
            "target_disp": target_disp,
            "target_vac": target_vac,
            "target_replac": target_replac,
            "target_inter": target_inter,
        }

    def _filter_subcollisions(
        self,
        srim_dir: Path,
        cur: sqlite3.Cursor,
        nions: Union[int, np.int64],
        trimdat: np.ndarray,
        nsubcollisions0: np.ndarray,
        criterion: Callable,
    ) -> tuple:
        """Filters subcollisions.

        Parameters
        ----------
        cur : sqlite3.Cursor
            Database cursor.
        nions : Union[int, np.int64]
            Number of ions.
        trimdat : dtypes.trimdat
            TRIMDAT data.
        nsubcollisions0 : dtypes.INT_CHECK
            Initial number of subcollisions.
        criterion : Callable
            Criterion to repeat calculation, must return False to repeat calculation.
        """
        self.append_subcollision(srim_dir / "SRIM Outputs/COLLISON.txt")
        atomic_numbers, recoil_energies, depths, ys, zs, cosxs, cosys, coszs = (
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
        )
        nsubcollisions = np.zeros(nions, dtype=int)
        for nion in range(nions):
            nsubcollision_min = 0 if nion == 0 else nsubcollision_max
            nsubcollision_max = nsubcollision_min + nsubcollisions0[nion]
            for nsubcollision in range(nsubcollision_min, nsubcollision_max):
                pos0 = trimdat[nsubcollision]["pos"]
                cosx0, cosy0, cosz0 = trimdat[nsubcollision]["dir"]
                for subcollision in self.subcollision.read(
                    condition=f"WHERE ion_numb = {nsubcollision + 1}"
                ):
                    subcollision = self._filter_subcollisions_logic(subcollision)
                    pos = np.array(
                        [subcollision["depth"], subcollision["y"], subcollision["z"]]
                    )
                    cosx, cosy, cosz = self._get_dir(pos0, pos)
                    # There are some rare cases, specially at high energies,
                    # when there are two PKA at the same position. I think this
                    # is because they are really close and when saved into
                    # COLLISON.txt, positions are rounded and they coincide.
                    # In such cases (if statement triggered), we assume that the
                    # second PKA has the same direction as the first one.
                    if np.isnan(cosx) or np.isnan(cosy) or np.isnan(cosz):
                        cosx, cosy, cosz = cosx0, cosy0, cosz0
                    else:
                        cosx0, cosy0, cosz0 = cosx, cosy, cosz
                    subcollision.update({"cosx": cosx, "cosy": cosy, "cosz": cosz})
                    # Check if SRIM has to be run again
                    ok = criterion(nion=nion + 1, **subcollision)
                    # If not, save into the database
                    if ok:
                        self.collision.insert(cur, nion + 1, **subcollision)
                    # Else, save it for later
                    else:
                        atomic_numbers.append(
                            materials.get_atomic_number_by_name(
                                subcollision["atom_hit"]
                            )
                        )
                        recoil_energies.append(subcollision["recoil_energy"])
                        depths.append(subcollision["depth"])
                        ys.append(subcollision["y"])
                        zs.append(subcollision["z"])
                        cosxs.append(subcollision["cosx"])
                        cosys.append(subcollision["cosy"])
                        coszs.append(subcollision["cosz"])
                        nsubcollisions[nion] += 1
                    pos0 = pos
                    # print(pos0, [cosx, cosy, cosz])
        self.subcollision.empty()
        atomic_numbers = np.array(atomic_numbers, dtype=int)
        recoil_energies, depths, ys, zs = (
            np.array(recoil_energies),
            np.array(depths),
            np.array(ys),
            np.array(zs),
        )
        cosxs, cosys, coszs = np.array(cosxs), np.array(cosys), np.array(coszs)
        return (
            nsubcollisions,
            atomic_numbers,
            recoil_energies,
            depths,
            ys,
            zs,
            cosxs,
            cosys,
            coszs,
        )

    def __remove_mean_depth_offsets(
        self, cur: sqlite3.Cursor, table_name: str, depth_mean: tuple[float]
    ) -> None:
        """Removes ion mean depth offsets from the given table.

        Parameters
        ----------
        cur : sqlite3.Cursor
            Database cursor.
        table_name : str
            Table name.
        depth_mean : tuple[float]
            Mean depth.
        """
        cur.execute(f"UPDATE {table_name} SET depth = depth - ?", depth_mean)

    def __remove_individual_offsets(self, cur: sqlite3.Cursor, table_name: str) -> None:
        """Removes individual ion offsets from the given table.

        Parameters
        ----------
        cur : sqlite3.Cursor
            Database cursor.
        table_name : str
            Table name.
        """
        cur.execute(
            f"""
            UPDATE {table_name}
            SET depth = {table_name}.depth - trimdat.depth,
                y = {table_name}.y - trimdat.y,
                z = {table_name}.z - trimdat.z
            FROM trimdat
            WHERE {table_name}.ion_numb = trimdat.ion_numb
        """
        )

    def __remove_offsets(self) -> None:
        """Removes ion offsets."""
        cur = self.cursor()
        cur.execute("SELECT AVG(depth) FROM trimdat")
        depth_mean = cur.fetchone()
        self.__remove_mean_depth_offsets(cur, "e2recoil", depth_mean)
        self.__remove_mean_depth_offsets(cur, "ioniz", depth_mean)
        self.__remove_mean_depth_offsets(cur, "lateral", depth_mean)
        self.__remove_mean_depth_offsets(cur, "phonon", depth_mean)
        self.__remove_mean_depth_offsets(cur, "range", depth_mean)
        self.__remove_mean_depth_offsets(cur, "vacancy", depth_mean)
        if self.calculation in ["full", "mono"]:
            self.__remove_mean_depth_offsets(cur, "novac", depth_mean)
        self.__remove_individual_offsets(cur, "backscat")
        self.__remove_individual_offsets(cur, "collision")
        self.__remove_individual_offsets(cur, "range3d")
        self.__remove_individual_offsets(cur, "sputter")
        self.__remove_individual_offsets(cur, "transmit")
        self.__remove_individual_offsets(cur, "transmit")
        cur.close()
        self.commit()



"""This module contains the `Collision` class."""

import sqlite3
from typing import TYPE_CHECKING, Generator

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Collision(SRIMFile):
    """Class for processing the `COLLISON.txt` file."""

    def __init__(self, srimdb: "SRIMDB") -> None:
        """Initializes the `Collision` object.

        Parameters
        ----------
        srimdb : SRIMDB
            `SRIMDB` object.
        """
        super().__init__(srimdb)
        if self.srim.calculation == "quick":
            self.create_table = self.__create_table_qc
            self.insert = self.__insert_qc
            self.merge = self.__merge_qc
        else:
            self.create_table = self.__create_table_fc
            self.insert = self.__insert_fc
            self.merge = self.__merge_fc

    def __create_table_qc(self) -> None:
        """Creates the collision table for Quick-Calculation mode."""
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE collision"
                "(ion_numb INTEGER, energy REAL, depth REAL, y REAL, z REAL,"
                "cosx REAL, cosy REAL, cosz REAL, se REAL, atom_hit TEXT,"
                "recoil_energy REAL, target_disp REAL)"
            )
        )
        cur.close()
        self.srim.commit()

    def __create_table_fc(self) -> None:
        """Creates the collision table for Full-Calculation mode."""
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE collision"
                "(ion_numb INTEGER, energy REAL, depth REAL, y REAL, z REAL,"
                "cosx REAL, cosy REAL, cosz REAL, se REAL, atom_hit TEXT,"
                "recoil_energy REAL, target_disp INTEGER, target_vac INTEGER,"
                "target_replac INTEGER, target_inter INTEGER)"
            )
        )
        cur.close()
        self.srim.commit()

    def __insert_qc(
        self,
        cur: sqlite3.Cursor,
        nion: int,
        energy: float,
        depth: float,
        y: float,
        z: float,
        cosx: float,
        cosy: float,
        cosz: float,
        se: float,
        atom_hit: str,
        recoil_energy: float,
        target_disp: float,
    ) -> None:
        """Inserts data into the collision table for Quick-Calculation mode.

        Parameters
        ----------
        cur : sqlite3.Cursor
            Database cursor.
        nion : int
            Ion number.
        energy : float
            Energy.
        depth : float
            Depth.
        y : float
            Y coordinate.
        z : float
            Z coordinate.
        cosx : float
            X direction.
        cosy : float
            Y direction.
        cosz : float
            Z direction.
        se : float
            Stopping power.
        atom_hit : str
            Atom hit.
        recoil_energy : float
            Recoil energy.
        target_disp : int
            Target displacement.
        """
        # try:
        cur.execute(
            (
                "INSERT INTO collision"
                "(ion_numb, energy, depth, y, z, cosx, cosy, cosz, se, atom_hit,"
                "recoil_energy, target_disp)"
                "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
            ),
            [
                nion,
                energy,
                depth,
                y,
                z,
                cosx,
                cosy,
                cosz,
                se,
                atom_hit,
                recoil_energy,
                target_disp,
            ],
        )

    def __insert_fc(
        self,
        cur: sqlite3.Cursor,
        nion: int,
        energy: float,
        depth: float,
        y: float,
        z: float,
        cosx: float,
        cosy: float,
        cosz: float,
        se: float,
        atom_hit: str,
        recoil_energy: float,
        target_disp: int,
        target_vac: int,
        target_replac: int,
        target_inter: int,
    ) -> None:
        """Inserts data into the collision table for Full-Calculation mode.

        Parameters
        ----------
        cur : sqlite3.Cursor
            Database cursor.
        nion : int
            Ion number.
        energy : float
            Energy.
        depth : float
            Depth.
        y : float
            Y coordinate.
        z : float
            Z coordinate.
        cosx : float
            X direction.
        cosy : float
            Y direction.
        cosz : float
            Z direction.
        se : float
            Stopping power.
        atom_hit : str
            Atom hit.
        recoil_energy : float
            Recoil energy.
        target_disp : int
            Target displacement.
        target_vac : int
            Target vacancy.
        target_replac : int
            Target replacement.
        target_inter : int
            Target interstitial.
        """
        # try:
        cur.execute(
            (
                "INSERT INTO collision"
                "(ion_numb, energy, depth, y, z, cosx, cosy, cosz, se, atom_hit,"
                "recoil_energy, target_disp, target_vac, target_replac,"
                "target_inter)"
                "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
            ),
            [
                nion,
                energy,
                depth,
                y,
                z,
                cosx,
                cosy,
                cosz,
                se,
                atom_hit,
                recoil_energy,
                target_disp,
                target_vac,
                target_replac,
                target_inter,
            ],
        )

    def __merge_qc(self, srimdb2: "SRIMDB") -> None:
        """Merges the collision table with another database for Quick-Calculation mode.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions = self.srim.nions
        cur = self.cursor()
        cur.execute(f"ATTACH DATABASE '{srimdb2.db_path}' AS srimdb2")
        cur.execute(
            (
                "INSERT INTO collision"
                "(ion_numb, energy, depth, y, z, cosx, cosy, cosz, se, atom_hit,"
                "recoil_energy, target_disp, target_vac, target_replac,"
                "target_inter) SELECT ion_numb + ?, energy, depth, y, z, cosx, cosy,"
                "cosz, se, atom_hit, recoil_energy, target_disp "
                "FROM srimdb2.collision"
            ),
            (nions,),
        )
        self.srim.commit()
        cur.execute("DETACH DATABASE srimdb2")
        cur.close()

    def __merge_fc(self, srimdb2: "SRIMDB") -> None:
        """Merges the collision table with another database for Full-Calculation mode.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions = self.srim.nions
        cur = self.cursor()
        cur.execute(f"ATTACH DATABASE '{srimdb2.db_path}' AS srimdb2")
        cur.execute(
            (
                "INSERT INTO collision"
                "(ion_numb, energy, depth, y, z, cosx, cosy, cosz, se, atom_hit,"
                "recoil_energy, target_disp, target_vac, target_replac,"
                "target_inter) SELECT ion_numb + ?, energy, depth, y, z, cosx, cosy,"
                "cosz, se, atom_hit, recoil_energy, target_disp, target_vac,"
                "target_replac, target_inter FROM srimdb2.collision"
            ),
            (nions,),
        )
        self.srim.commit()
        cur.execute("DETACH DATABASE srimdb2")
        cur.close()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads data from the collision table.

        Parameters
        ----------
        what : str, optional
            Columns to read, by default "*".
        condition : str, optional
            Condition to filter data, by default "".

        Yields
        ------
        tuple
            Data from the collision table.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM collision {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()



"""This module contains the `Phonon` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Phonon(SRIMFile):
    """Class to handle `PHONON.txt` file."""

    def process_file(self, phonon_path: Path) -> None:
        """Processes `PHONON.txt` file.

        Parameters
        ----------
        phonon_path : Path
            `PHONON.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            (
                "CREATE TABLE phonon"
                "(depth REAL, phonons_ions REAL, phonons_recoils REAL)"
            )
        )
        with open(phonon_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("  DEPTH"):
                    break
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                phonons_ions = data[1]
                phonons_recoils = data[2]
                cur.execute(
                    (
                        "INSERT INTO phonon(depth, phonons_ions, phonons_recoils)"
                        "VALUES(?, ?, ?)"
                    ),
                    [depth, phonons_ions, phonons_recoils],
                )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads phonon data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM phonon {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the phonon table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        phonon1 = np.array(list(self.read()))
        phonon2 = np.array(list(srimdb2.phonon.read()))
        phonon1[:, 1:] *= nions1
        phonon2[:, 1:] *= nions2
        phonon1[:, 1:] += phonon2[:, 1:]
        phonon1[:, 1:] /= nions
        cur = self.cursor()
        cur.execute("DELETE FROM phonon")
        cur.executemany(
            (
                "INSERT INTO phonon(depth, phonons_ions, phonons_recoils)"
                "VALUES(?, ?, ?)"
            ),
            phonon1,
        )
        cur.close()
        self.srim.commit()



"""This module contains the `LAMMPSReader` class."""

from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Type, Union

import numpy as np


@dataclass
class LAMMPSReader:
    """A class to read data from a LAMMPS dump file.

    Note
    ----
    Assumed orthogonal simulation box.

    Attributes
    ----------
    file_path : Path
        The path to the LAMMPS dump file.
    """

    file_path: Path

    def __get_dtype(
        self, line: str
    ) -> tuple[list[str], list[Type[Union[int, float]]], np.dtype]:
        """Get the data type of the simulation data.

        Parameters
        ----------
        line : str
            The line containing the data type.

        Returns
        -------
        tuple[list[str], list[Type[Union[int, float]], np.dtype]
            The names of the data items, the types of the data items,
            and the data type.
        """
        items = line.split()[2:]
        types = [int if item in ["id", "type", "element"] else float for item in items]
        dtype = np.dtype([(item, type) for item, type in zip(items, types)])
        return items, types, dtype

    def __iter__(
        self,
    ) -> Generator[
        tuple[
            Union[float, None],
            int,
            int,
            tuple[str, str, str],
            tuple[float, float],
            tuple[float, float],
            tuple[float, float],
            np.ndarray,
        ],
        None,
        None,
    ]:
        """Read the file as an iterator, timestep by timestep.

        Yields
        ------
        tuple[int, np.ndarray]
            The timestep and the data at that timestep.
        """
        with open(self.file_path, encoding="utf-8") as file:
            while True:
                line = file.readline()
                if not line:
                    break
                time = None
                if line == "ITEM: TIME\n":
                    time = float(file.readline())
                    file.readline()
                timestep = int(file.readline())
                file.readline()
                natoms = int(file.readline())
                boundary = list(file.readline().split()[-3:])
                boundary = (boundary[0][0], boundary[1][0], boundary[2][0])
                dimsx = tuple(map(float, file.readline().split()))
                dimsy = tuple(map(float, file.readline().split()))
                dimsz = tuple(map(float, file.readline().split()))

                line = file.readline()
                items, types, dtype = self.__get_dtype(line)
                data = np.empty(natoms, dtype=dtype)
                for i in range(natoms):
                    line = file.readline().split()
                    for j, item in enumerate(items):
                        data[i][item] = types[j](line[j])
                yield time, timestep, natoms, boundary, dimsx, dimsy, dimsz, data



"""This module contains the `Ioniz` class."""

from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np

from irradiapy.srimpy.ofiles.srimfile import SRIMFile

if TYPE_CHECKING:
    from irradiapy.srimpy.srimdb import SRIMDB


class Ioniz(SRIMFile):
    """Class to handle `IONIZ.txt` file."""

    def process_file(self, ioniz_path: Path) -> None:
        """Processes `IONIZ.txt` file.

        Parameters
        ----------
        ioniz_path : Path
            `IONIZ.txt` path.
        """
        cur = self.cursor()
        cur.execute(
            "CREATE TABLE ioniz (depth REAL, ioniz_ions REAL, ioniz_recoils REAL)"
        )
        with open(ioniz_path, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith("  TARGET"):
                    break
            next(file)
            next(file)
            next(file)
            for _ in range(100):
                line = next(file)
                data = list(map(float, line[:-1].split()))
                depth = data[0]
                ioniz_ions = data[1]
                ioniz_recoils = data[2]
                cur.execute(
                    (
                        "INSERT INTO ioniz(depth, ioniz_ions, ioniz_recoils)"
                        "VALUES(?, ?, ?)"
                    ),
                    [depth, ioniz_ions, ioniz_recoils],
                )
        cur.close()
        self.srim.commit()

    def read(
        self, what: str = "*", condition: str = ""
    ) -> Generator[tuple, None, None]:
        """Reads ioniz data from the database as a generator.

        Parameters
        ----------
        what : str
            Columns to select.
        condition : str
            Condition to filter data.

        Yields
        ------
        Generator[tuple, None, None]
            Data from the database.
        """
        cur = self.cursor()
        cur.execute(f"SELECT {what} FROM ioniz {condition}")
        while True:
            data = cur.fetchone()
            if data:
                yield data
            else:
                break
        cur.close()

    def merge(self, srimdb2: "SRIMDB") -> None:
        """Merges the ioniz table with another database.

        Parameters
        ----------
        srimdb2 : SRIMDB
            SRIM database to merge.
        """
        nions1 = self.srim.nions
        nions2 = srimdb2.nions
        nions = nions1 + nions2
        ioniz1 = np.array(list(self.read()))
        ioniz2 = np.array(list(srimdb2.ioniz.read()))
        ioniz1[:, 1:] *= nions1
        ioniz2[:, 1:] *= nions2
        ioniz1[:, 1:] += ioniz2[:, 1:]
        ioniz1[:, 1:] /= nions
        cur = self.cursor()
        cur.execute("DELETE FROM ioniz")
        cur.executemany(
            "INSERT INTO ioniz(depth, ioniz_ions, ioniz_recoils) VALUES(?, ?, ?)",
            ioniz1,
        )
        cur.close()
        self.srim.commit()



"""This module contains the `LAMMPSWriter` class."""

from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import Optional


@dataclass
class LAMMPSWriter:
    """A class to write data like a LAMMPS dump file.

    Attributes
    ----------
    file_path : Path
        The path to the LAMMPS dump file.
    """

    file_path: Path
    mode: str
    encoding: str = "utf-8"

    def __post_init__(self) -> None:
        """Opens the file associated with this writer."""
        self.file = open(self.file_path, self.mode, encoding=self.encoding)

    def __enter__(self) -> "LAMMPSWriter":
        """Enters the context manager."""
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        exc_traceback: Optional[TracebackType] = None,
    ) -> bool:
        """Exits the context manager."""
        self.file.close()
        return False

    def close(self) -> None:
        """Closes the file associated with this writer."""
        self.file.close()

    def __del__(self) -> None:
        """Closes the file associated with this writer."""
        self.file.close()

    def save(
        self,
        timestep: int,
        boundary: tuple[str, str, str],
        dimsx: tuple[float, float],
        dimsy: tuple[float, float],
        dimsz: tuple[float, float],
        time: Optional[float] = None,
        **kwargs,
    ) -> None:
        """Writes the data to the file.

        Note
        ----
        Assumes orthogonal simulation box.

        Parameters
        ----------
        timestep : int
            The timestep of the simulation.
        boundary : tuple[str, str, str]
            The type of boundary conditions of the simulation.
        dimsx : tuple[float, float]
            The maximum extents of the simulation box in the x-dimension,
        dimsy : tuple[float, float]
            The maximum extents of the simulation box in the y-dimension.
        dimsz : tuple[float, float]
            The maximum extents of the simulation box in the z-dimension.
        time : float, optional
            The time of the simulation.
        **kwargs
            The data to be written to the file.
        """
        field_names = list(kwargs.keys())
        natoms = len(kwargs[field_names[0]])
        if time is not None:
            self.file.write(f"ITEM: TIME\n{time}\n")
        self.file.write(f"ITEM: TIMESTEP\n{timestep}\n")
        self.file.write(f"ITEM: NUMBER OF ATOMS\n{natoms}\n")
        self.file.write(
            f"ITEM: BOX BOUNDS {boundary[0]*2} {boundary[1]*2} {boundary[2]*2}\n"
        )
        self.file.write(f"{dimsx[0]} {dimsx[1]}\n")
        self.file.write(f"{dimsy[0]} {dimsy[1]}\n")
        self.file.write(f"{dimsz[0]} {dimsz[1]}\n")
        self.file.write(f"ITEM: ATOMS {' '.join(field_names)}\n")
        for i in range(natoms):
            row = [kwargs[field][i] for field in field_names]
            self.file.write(" ".join(map(str, row)) + "\n")



"""This module contains the `Element` class."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class Element:
    """Class for defining an element in SRIM simulations.

    Attributes
    ----------
    symbol : str
        The symbol of the element.
    atomic_number : int
        The atomic number of the element.
    atomic_mass : float
        The atomic mass of the element in atomic mass units.
    e_d : float
        The displacement energy of the element in eV.
    e_l : float
        The lattice binding energy of the element in eV.
    e_s : float
        The surface binding energy of the element in eV.
    density : float, optional
        The density of the element in g/cm^3. Only to be used by predefined
        materials to simplify the definition of the layers in the target by the
        user.
    """

    symbol: str
    atomic_number: int
    atomic_mass: float
    e_d: float
    e_l: float
    e_s: float
    density: Optional[float] = None


# Some predefined elements
# Check this reference for recommended values:
# https://doi.org/10.1016/j.nimb.2021.06.018
#: Element: Predefined chromium.
Cr = Element("Cr", 24, 51.9961, 40.0, 7.8, 13.2, 7.19)
#: Element: Predefined iron.
Fe = Element("Fe", 26, 55.85, 40.0, 5.8, 4.34, 7.8658)
#: Element: Predefined silver.
Ag = Element("Ag", 47, 107.8682, 39.0, 4.0, 2.97, 10.49)
#: Element: Predefined tungsten.
W = Element("W", 74, 183.84, 70.0, 13.2, 8.68, 19.3)
#: Element: Predefined copper.
Cu = Element("Cu", 29, 63.546, 33.0, 4.4, 3.52, 8.92)



"""This module contains the `Material` class and some predefined materials."""

from dataclasses import dataclass
from math import sqrt
from typing import Optional


@dataclass
class Material:
    """Class for storing parameters of a material.

    Parameters
    ----------
    atomic_number : int
        Atomic number.
    mass_number : float
        Mass number (atomic mass units).
    a0 : float, optional
        Lattice parameter (Å). Default is None.
    cutoff_sia : float, optional
        Cutoff distance for interstitial clusters detection (Å). Default is None.
    cutoff_vac : float, optional
        Cutoff distance for vacancy clusters detection (Å). Default is None.
    dist_fp : float, optional
        Frenkel pair distance (Å). Default is None.
    density : float, optional
        Atomic density (atoms/Å³). Default is None.
    ed_min : float, optional
        Minimum displacement energy (eV). Default is None.
    ed_avr : float, optional
        Average displacement energy (eV). Default is None.
    b_arc : float, optional
        'b' parameter of the arc-dpa fit. Default is None.
    c_arc : float, optional
        'c' parameter of the arc-dpa fit. Default is None.
    """

    atomic_number: int
    mass_number: float
    a0: Optional[float] = None
    cutoff_sia: Optional[float] = None
    cutoff_vac: Optional[float] = None
    dist_fp: Optional[float] = None
    density: Optional[float] = None
    ed_min: Optional[float] = None
    ed_avr: Optional[float] = None
    b_arc: Optional[float] = None
    c_arc: Optional[float] = None


def get_material_by_name(name: str) -> Material:
    """Retrieve a predefined Material by its name.

    Parameters
    ----------
    name : str
        The name of the predefined material.

    Returns
    -------
    Material
        The requested Material instance.

    Raises
    ------
    AttributeError
        If the material is not defined.
    """
    material = PREDEFINED_MATERIALS.get(name)
    if material is None:
        err_msg = f"Material with name '{name}' is not defined."
        raise AttributeError(err_msg)
    return material


def get_material_by_atomic_number(atomic_number: int) -> Material:
    """Retrieve a predefined Material by its atomic number.

    Parameters
    ----------
    atomic_number : int
        The atomic number of the material.

    Returns
    -------
    Material
        The requested Material instance.

    Raises
    ------
    AttributeError
        If the material is not defined.
    """
    material = MATERIALS_BY_ATOMIC_NUMBER.get(atomic_number)
    if material is None:
        err_msg = f"Material with atomic number {atomic_number} is not defined."
        raise AttributeError(err_msg)
    return material


def get_atomic_number_by_name(name: str) -> int:
    """Retrieve the atomic number of a material by its name.

    Parameters
    ----------
    name : str
        The name of the material.

    Returns
    -------
    int
        The atomic number of the material.

    Raises
    ------
    AttributeError
        If the material is not defined.
    """
    material = PREDEFINED_MATERIALS.get(name)
    if material is None:
        err_msg = f"Material with name '{name}' is not defined."
        raise AttributeError(err_msg)
    return material.atomic_number


def get_mass_number_by_name(name: str) -> float:
    """Retrieve the mass number of a material by its name.

    Parameters
    ----------
    name : str
        The name of the material.

    Returns
    -------
    float
        The mass number of the material.

    Raises
    ------
    AttributeError
        If the material is not defined.
    """
    material = PREDEFINED_MATERIALS.get(name)
    if material is None:
        err_msg = f"Material with name '{name}' is not defined."
        raise AttributeError(err_msg)
    return material.mass_number


def get_mass_number_by_atomic_number(atomic_number: int) -> float:
    """Retrieve the mass number of a material by its atomic number.

    Parameters
    ----------
    atomic_number : int
        The atomic number of the material.

    Returns
    -------
    float
        The mass number of the material.

    Raises
    ------
    AttributeError
        If the material is not defined.
    """
    material = MATERIALS_BY_ATOMIC_NUMBER.get(atomic_number)
    if material is None:
        err_msg = f"Material with atomic number {atomic_number} is not defined."
        raise AttributeError(err_msg)
    return material.mass_number


#: Material: Predefined materials.
PREDEFINED_MATERIALS = {
    "H": Material(atomic_number=1, mass_number=1.008),
    "He": Material(atomic_number=2, mass_number=4.002602),
    "Li": Material(atomic_number=3, mass_number=6.94),
    "Be": Material(atomic_number=4, mass_number=9.0122),
    "B": Material(atomic_number=5, mass_number=10.81),
    "C": Material(atomic_number=6, mass_number=12.011),
    "N": Material(atomic_number=7, mass_number=14.007),
    "O": Material(atomic_number=8, mass_number=15.999),
    "F": Material(atomic_number=9, mass_number=18.998),
    "Ne": Material(atomic_number=10, mass_number=20.180),
    "Na": Material(atomic_number=11, mass_number=22.990),
    "Mg": Material(atomic_number=12, mass_number=24.305),
    "Al": Material(atomic_number=13, mass_number=26.982),
    "Si": Material(atomic_number=14, mass_number=28.085),
    "P": Material(atomic_number=15, mass_number=30.974),
    "S": Material(atomic_number=16, mass_number=32.06),
    "Cl": Material(atomic_number=17, mass_number=35.45),
    "Ar": Material(atomic_number=18, mass_number=39.948),
    "K": Material(atomic_number=19, mass_number=39.098),
    "Ca": Material(atomic_number=20, mass_number=40.078),
    "Sc": Material(atomic_number=21, mass_number=44.956),
    "Ti": Material(atomic_number=22, mass_number=47.867),
    "V": Material(atomic_number=23, mass_number=50.942),
    "Cr": Material(atomic_number=24, mass_number=51.996),
    "Mn": Material(atomic_number=25, mass_number=54.938),
    "Fe": Material(
        atomic_number=26,
        mass_number=55.845,
        a0=2.87,
        cutoff_sia=2.87 * sqrt(2.0),
        cutoff_vac=2.87,
        dist_fp=4.0 * 2.87,
        density=8.5e-2,
        ed_min=20,
        ed_avr=40,
        b_arc=-0.568,
        c_arc=0.286,
    ),
    "Co": Material(atomic_number=27, mass_number=58.933),
    "Ni": Material(atomic_number=28, mass_number=58.693),
    "Cu": Material(
        atomic_number=29,
        mass_number=63.546,
        a0=3.61491,
        cutoff_sia=3.61491 * sqrt(2.0),
        cutoff_vac=3.61491,
        dist_fp=4.0 * 3.61491,
        density=8.46e-2,
        ed_min=25,
        ed_avr=33,
        b_arc=-0.68,
        c_arc=0.16,
    ),
    "Zn": Material(atomic_number=30, mass_number=65.38),
    "Ga": Material(atomic_number=31, mass_number=69.723),
    "Ge": Material(atomic_number=32, mass_number=72.63),
    "As": Material(atomic_number=33, mass_number=74.921595),
    "Se": Material(atomic_number=34, mass_number=78.971),
    "Br": Material(atomic_number=35, mass_number=79.904),
    "Kr": Material(atomic_number=36, mass_number=83.798),
    "Rb": Material(atomic_number=37, mass_number=85.468),
    "Sr": Material(atomic_number=38, mass_number=87.62),
    "Y": Material(atomic_number=39, mass_number=88.90584),
    "Zr": Material(atomic_number=40, mass_number=91.224),
    "Nb": Material(atomic_number=41, mass_number=92.906),
    "Mo": Material(atomic_number=42, mass_number=95.95),
    "Tc": Material(atomic_number=43, mass_number=98),
    "Ru": Material(atomic_number=44, mass_number=101.07),
    "Rh": Material(atomic_number=45, mass_number=102.91),
    "Pd": Material(atomic_number=46, mass_number=106.42),
    "Ag": Material(atomic_number=47, mass_number=107.87),
    "Cd": Material(atomic_number=48, mass_number=112.41),
    "In": Material(atomic_number=49, mass_number=114.82),
    "Sn": Material(atomic_number=50, mass_number=118.71),
    "Sb": Material(atomic_number=51, mass_number=121.760),
    "Te": Material(atomic_number=52, mass_number=127.60),
    "I": Material(atomic_number=53, mass_number=126.90447),
    "Xe": Material(atomic_number=54, mass_number=131.29),
    "Cs": Material(atomic_number=55, mass_number=132.91),
    "Ba": Material(atomic_number=56, mass_number=137.33),
    "La": Material(atomic_number=57, mass_number=138.91),
    "Ce": Material(atomic_number=58, mass_number=140.12),
    "Pr": Material(atomic_number=59, mass_number=140.91),
    "Nd": Material(atomic_number=60, mass_number=144.242),
    "Pm": Material(atomic_number=61, mass_number=145),
    "Sm": Material(atomic_number=62, mass_number=150.36),
    "Eu": Material(atomic_number=63, mass_number=151.964),
    "Gd": Material(atomic_number=64, mass_number=157.25),
    "Tb": Material(atomic_number=65, mass_number=158.92535),
    "Dy": Material(atomic_number=66, mass_number=162.500),
    "Ho": Material(atomic_number=67, mass_number=164.93033),
    "Er": Material(atomic_number=68, mass_number=167.259),
    "Tm": Material(atomic_number=69, mass_number=168.93422),
    "Yb": Material(atomic_number=70, mass_number=173.04),
    "Lu": Material(atomic_number=71, mass_number=174.9668),
    "Hf": Material(atomic_number=72, mass_number=178.49),
    "Ta": Material(atomic_number=73, mass_number=180.94788),
    "W": Material(
        atomic_number=74,
        mass_number=183.84,
        a0=3.1652,
        cutoff_sia=3.1652 * sqrt(2.0),
        cutoff_vac=3.1652,
        dist_fp=4.0 * 3.1652,
        density=6.3e-2,
        ed_min=42,
        ed_avr=70,
        b_arc=-0.56,
        c_arc=0.12,
    ),
    "Re": Material(atomic_number=75, mass_number=186.21),
    "Os": Material(atomic_number=76, mass_number=190.23),
    "Ir": Material(atomic_number=77, mass_number=192.22),
    "Pt": Material(atomic_number=78, mass_number=195.084),
    "Au": Material(atomic_number=79, mass_number=196.97),
    "Hg": Material(atomic_number=80, mass_number=200.592),
    "Tl": Material(atomic_number=81, mass_number=204.38),
    "Pb": Material(atomic_number=82, mass_number=207.2),
    "Bi": Material(atomic_number=83, mass_number=208.98040),
    "Po": Material(atomic_number=84, mass_number=209),
    "At": Material(atomic_number=85, mass_number=210),
    "Rn": Material(atomic_number=86, mass_number=222),
    "Fr": Material(atomic_number=87, mass_number=223),
    "Ra": Material(atomic_number=88, mass_number=226),
    "Ac": Material(atomic_number=89, mass_number=227),
    "Th": Material(atomic_number=90, mass_number=232.03805),
    "Pa": Material(atomic_number=91, mass_number=231.03588),
    "U": Material(atomic_number=92, mass_number=238.02891),
    "Np": Material(atomic_number=93, mass_number=237),
    "Pu": Material(atomic_number=94, mass_number=244),
    "Am": Material(atomic_number=95, mass_number=243),
    "Cm": Material(atomic_number=96, mass_number=247),
    "Bk": Material(atomic_number=97, mass_number=247),
    "Cf": Material(atomic_number=98, mass_number=251),
    "Es": Material(atomic_number=99, mass_number=252),
    "Fm": Material(atomic_number=100, mass_number=257),
    "Md": Material(atomic_number=101, mass_number=258),
    "No": Material(atomic_number=102, mass_number=259),
    "Lr": Material(atomic_number=103, mass_number=262),
    "Rf": Material(atomic_number=104, mass_number=267),
    "Db": Material(atomic_number=105, mass_number=270),
    "Sg": Material(atomic_number=106, mass_number=271),
    "Bh": Material(atomic_number=107, mass_number=270),
    "Hs": Material(atomic_number=108, mass_number=277),
    "Mt": Material(atomic_number=109, mass_number=278),
    "Ds": Material(atomic_number=110, mass_number=281),
    "Rg": Material(atomic_number=111, mass_number=282),
    "Cn": Material(atomic_number=112, mass_number=285),
    "Nh": Material(atomic_number=113, mass_number=286),
    "Fl": Material(atomic_number=114, mass_number=289),
    "Mc": Material(atomic_number=115, mass_number=290),
    "Lv": Material(atomic_number=116, mass_number=293),
    "Ts": Material(atomic_number=117, mass_number=294),
    "Og": Material(atomic_number=118, mass_number=294),
}
MATERIALS_BY_ATOMIC_NUMBER = {
    mat.atomic_number: mat for mat in PREDEFINED_MATERIALS.values()
}



"""Utility functions related to dpa for SRIM data."""

import sqlite3
from pathlib import Path
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D

from irradiapy import dpa, materials
from irradiapy.io.xyzreader import XYZReader
from irradiapy.math_utils import fit_lorentzian
from irradiapy.srimpy.srimdb import SRIMDB


def get_dpas(
    srimdb: SRIMDB,
    path_collisions: Path,
    fluence: float,
    path_db: Path,
    nbins: int = 100,
    depth_offset: Union[int, float, np.number] = 0.0,
) -> None:
    """Create a table with NRT, fer-arc and debris-dpa in a database.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    path_collisions : Path
        File containing all collision debris to analyze.
    fluence : float
        Fluence, in ions/A2.
    path_db : Path
        Path to the database file.
    nbins : int, optional
        Number of depth bins. Default is 100.
    """
    mat1 = materials.get_material_by_atomic_number(
        list(srimdb.trimdat.read(what="atom_numb", condition="WHERE ion_numb = 1"))[0][
            0
        ]
    )
    mat2 = materials.get_material_by_atomic_number(
        srimdb.target.layers[0].elements[0].atomic_number
    )
    nions = srimdb.nions
    # SRIM COLLISON.txt dpa
    depths, defects_nrt, defects_fer_arc = [], [], []
    for depth, pka_e in srimdb.collision.read(what="depth, recoil_energy"):
        depths.append(depth + depth_offset)
        tdam = dpa.compute_damage_energy(pka_e, mat1, mat2)
        defects_nrt.append(dpa.calc_nrt_dpa(tdam, mat2))
        defects_fer_arc.append(dpa.calc_fer_arc_dpa(tdam, mat2))
    depths = np.array(depths)
    defects_nrt = np.array(defects_nrt)
    defects_fer_arc = np.array(defects_fer_arc)
    # Debris dpa
    depth_debris = np.array([], dtype=float)
    nion = 0
    total_debris = 0
    for defects in XYZReader(path_collisions):
        vacs = defects[defects["type"] == 0]
        depth_debris = np.concatenate((depth_debris, vacs["pos"][:, 0]))
        nion += 1
        total_debris += vacs.size
        if nion % 100 == 0:
            print(f"{nion}/{nions}")
    # Depth binning
    depth_edges = np.histogram_bin_edges(depths, bins=nbins)
    depth_centers = (depth_edges[1:] - depth_edges[:-1]) / 2.0 + depth_edges[:-1]
    width = depth_edges[1] - depth_edges[0]
    depth_digitize = np.digitize(depths, depth_edges)
    # NRT-dpa
    hist_nrt = np.array(
        [np.sum(defects_nrt[depth_digitize == i]) for i in range(1, nbins + 1)]
    )
    dpa_nrt = hist_nrt / nions * fluence / mat2.density / width
    # fer-arc-dpa
    hist_fer_arc = np.array(
        [np.sum(defects_fer_arc[depth_digitize == i]) for i in range(1, nbins + 1)]
    )
    dpa_fer_arc = hist_fer_arc / nions * fluence / mat2.density / width
    # debris-dpa
    hist_debris, _ = np.histogram(depth_debris, bins=depth_edges)
    dpa_debris = hist_debris / nions * fluence / mat2.density / width
    # Save
    con = sqlite3.connect(path_db)
    cur = con.cursor()
    cur.execute("DROP TABLE IF EXISTS dpa")
    cur.execute("CREATE TABLE dpa (depth REAL, nrt REAL, arc REAL, debris REAL)")
    cur.executemany(
        "INSERT INTO dpa (depth, nrt, arc, debris) VALUES (?, ?, ?, ?)",
        np.column_stack((depth_centers, dpa_nrt, dpa_fer_arc, dpa_debris)),
    )
    con.commit()
    cur.close()
    con.close()


def load_results(path_db: Path, what: str = "*", condition: str = "") -> tuple:
    """Load results from the table created by `get_dpas`.

    Parameters
    ----------
    path_db : Path
        Path to the database file.
    what : str, optional
        Columns to select from the database, by default "*".
    condition : str, optional
        SQL condition for the query, by default "".

    Returns
    -------
    tuple
        Tuple containing depth_centers, dpa_nrt, dpa_fer_arc, and dpa_debris arrays.
    """
    con = sqlite3.connect(path_db)
    cur = con.cursor()
    cur.execute(f"SELECT {what} FROM dpa {condition}")
    data = np.array(cur.fetchall())
    depth_centers = data[:, 0]
    dpa_nrt = data[:, 1]
    dpa_fer_arc = data[:, 2]
    dpa_debris = data[:, 3]
    cur.close()
    con.close()
    return depth_centers, dpa_nrt, dpa_fer_arc, dpa_debris


def plot_dpa(
    path_db: Path,
    # depth_offset: Union[int, float, np.number] = 0.0,
    path_plot: Optional[Path] = None,
    dpi: int = 300,
    path_fit: Optional[Path] = None,
    p0: Optional[float] = None,
    asymmetry: float = 1.0,
) -> None:
    """Plot the dpa analysis results from the database.

    Parameters
    ----------
    path_db : Path
        Path to the database file.
    path_plot : Path, optional
        Output path for the plot, by default None. If None, the plot is shown.
    depth_offset : Union[int, float, np.number], optional
        Depth offset to apply to be applied, by default 0.0.
    dpi : int, optional
        Dots per inch, by default 300.
    path_fit : Path, optional
        Output path for the fit parameters, by default None.
    p0 : float, optional
        Initial guess of fit parameters, by default None.
    asymmetry : float, optional
        Asymmetry fit parameter bound, by default 1.0.
    """
    depth_centers, dpa_nrt, dpa_fer_arc, dpa_debris = load_results(path_db)
    # depth_centers += depth_offset
    total_nrt = dpa_nrt.sum()
    total_debris = dpa_debris.sum()
    # Fit dpa_debris
    fit = False
    if path_fit:
        try:
            # p0 = [depth_centers[np.argmax(dpa_debris)], 1.0, 1.0, 1.0]
            popt, _, dpa_fit = fit_lorentzian(depth_centers, dpa_debris, p0, asymmetry)
            if path_fit:
                with open(path_fit, "w", encoding="utf-8") as file:
                    file.write("Debris dpa lorentzian fit: z0, sigma, A, a\n")
                    file.write(
                        (
                            "See Eq. (1) of Nuclear Instruments and Methods in Physics "
                            "Research B 500-501 (2021) 52-56\n"
                        )
                    )
                    file.write(", ".join(map(str, popt)) + "\n")
            fit = True
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Fit failed: {exc}")
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec()
    ax = fig.add_subplot(gs[0, 0])
    # Scatter
    ax.set_xlabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_ylabel("dpa")
    ax.scatter(depth_centers, dpa_nrt, label="NRT-dpa")
    ax.scatter(depth_centers, dpa_fer_arc, label="fer-arc-dpa")
    ax.scatter(depth_centers, dpa_debris, label="debris-dpa")
    if fit:
        ax.plot(
            depth_centers,
            dpa_fit(depth_centers),
            label="debris-dpa fit",
        )
    efficiency = [
        Line2D(
            [0],
            [0],
            color="none",
            label=r"$\overline{\xi}$ = "
            + rf"{round(total_debris / total_nrt * 100)} %",
        )
    ]
    ax.legend(handles=ax.get_legend_handles_labels()[0] + efficiency)
    fig.tight_layout()
    if path_plot:
        plt.savefig(path_plot, dpi=dpi)
    else:
        plt.show()
    plt.close()



"""This subpackage provides a tool for irradiation damage analysis."""

from irradiapy.analysis import clusters
from irradiapy.analysis.defectsfinder import DefectsFinder
from irradiapy.analysis.defectsfinderold import DefectsFinderOld



"""."""

import numpy as np

field_type = ("type", int)
field_pos = ("pos", float, 3)
field_cluster = ("cluster", int)
field_size = ("size", int)
field_name = ("name", str)
field_atomic_number = ("atomic_number", int)
field_energy = ("energy", float)
field_dir = ("dir", float, 3)

atom = np.dtype([field_type, field_pos])
defect = np.dtype([field_type, field_pos])
acluster = np.dtype([field_pos, field_type, field_cluster])
ocluster = np.dtype([field_pos, field_type, field_size])
trimdat = np.dtype(
    [field_name, field_atomic_number, field_energy, field_pos, field_dir]
)
